# Copyright (c) 2024 Incorpo.ro. All rights reserved.
#
# This software is provided for non-commercial use only. Any commercial use, reproduction,
# distribution, or modification of this software, in whole or in part, is strictly prohibited
# without the prior written permission of Incorpo.ro.
#
# If you wish to use this software for any purpose, you must notify Incorpo.ro by email at
# office@incorpo.ro within 24 hours of starting to use the code.
#
# Unauthorized use of this software may result in a penalty, which shall be the highest of
# the following:
# - A fine of 1000 EUR
# - 20% of the profits made with the tool
# - The estimated increase in productivity generated by the use of this software
#
# By using this software, you agree to be bound by the terms of this copyright notice.
#
# For inquiries regarding commercial use or licensing, please contact Incorpo.ro at
# office@incorpo.ro.

import concurrent.futures
import anthropic
import base64
import json
import os

import requests
from bs4 import BeautifulSoup
from httpcore import TimeoutException
import time
from selenium import webdriver
from selenium.common import WebDriverException
from selenium.webdriver.chromium.options import ChromiumOptions
import xml.etree.ElementTree as ET
import re

from PIL import Image
from io import BytesIO
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By

client = anthropic.Anthropic(
    api_key="REDACTED",
)


def read_file(file_path):
    with open(file_path, "r") as file:
        return file.read()


def create_driver():
    options = ChromiumOptions()
    options.add_argument("--headless")  # Run Chrome in headless mode
    options.add_argument("--no-sandbox")  # Disable sandbox mode
    options.add_argument("--disable-dev-shm-usage")  # Disable /dev/shm usage
    return webdriver.Chrome(options=options)


MAX_RETRIES = 3
RETRY_DELAY = 5


def minify_and_clean_html(html):
    # Parse the HTML using BeautifulSoup
    soup = BeautifulSoup(html, 'html.parser')

    # Remove script and style tags
    for script in soup(["script", "style"]):
        script.decompose()

    # Remove unnecessary attributes from tags
    for tag in soup.find_all(True):
        tag.attrs = {}

    # Get the cleaned HTML string
    cleaned_html = str(soup)

    # Minify the HTML by removing extra whitespace
    minified_html = ' '.join(cleaned_html.split())

    return minified_html

def scrape_and_screenshot(driver, url):
    print(f"- Currently scraping: {url}")

    retries = 0
    while retries < MAX_RETRIES:
        try:
            driver.get(url)

            # Wait for a specific element to be present (adjust the selector as needed)
            wait = WebDriverWait(driver, 10)  # Timeout after 10 seconds
            wait.until(EC.presence_of_element_located((By.CSS_SELECTOR, "body")))

            # Set the viewport size to a large value to capture the entire page
            viewport_width = 1920
            viewport_height = 10000  # Adjust this value based on the page height
            driver.set_window_size(viewport_width, viewport_height)

            # Scrape the HTML
            html = minify_and_clean_html(driver.page_source)

            # Minify the html
            html = re.sub(r"\n", "", html)
            html = re.sub(r"\s+", " ", html)

            # Take only first 600k characters
            html = html[:450000]

            # Take a screenshot of the entire page
            screenshot_png = driver.get_screenshot_as_png()

            # Open the screenshot as an image
            screenshot = Image.open(BytesIO(screenshot_png))

            # Resize the image if it exceeds the maximum dimensions
            max_dimension = 6000
            width, height = screenshot.size
            if width > max_dimension or height > max_dimension:
                ratio = max_dimension / max(width, height)
                new_width = int(width * ratio)
                new_height = int(height * ratio)
                screenshot = screenshot.resize((new_width, new_height), Image.BICUBIC)

            # Convert the resized screenshot to JPEG with quality 90
            buffer = BytesIO()
            screenshot.convert('RGB').save(buffer, format='JPEG', quality=98)
            screenshot_jpeg_base64 = base64.b64encode(buffer.getvalue()).decode('utf-8')

            return html, screenshot_jpeg_base64

        except (TimeoutException, WebDriverException) as e:
            print(f"Error analyzing URL: {url}")
            print(f"Error message: {str(e)}")
            retries += 1
            print(f"Retrying ({retries}/{MAX_RETRIES}) in {RETRY_DELAY} seconds...")
            time.sleep(RETRY_DELAY)

    print(f"Max retries exceeded for URL: {url}")
    return None, None


def analyze_website(urls):
    with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        script_directory = os.path.dirname(os.path.abspath(__file__))
        system_prompt = read_file(f"{script_directory}/prompt/system-prompt.txt")

        content = [
            {
                "type": "text",
                "text": "<website_scrape>"
            }
        ]

        futures = []
        for url in urls:
            driver = create_driver()
            future = executor.submit(scrape_and_screenshot, driver, url)
            futures.append((future, driver, url))

        for future, driver, url in futures:
            try:
                html, screenshot = future.result()

                content.extend([
                    {
                        "type": "text",
                        "text": f"<page><url>{url}</url><html>{html}</html><page_render>"
                    },
                    {
                        "type": "image",
                        "source": {
                            "type": "base64",
                            "media_type": "image/jpeg",
                            "data": screenshot
                        }
                    },
                    {
                        "type": "text",
                        "text": "</page_render></page>"
                    }
                ])
            except Exception as e:
                print(f"Error analyzing URL: {url}")
                print(f"Error message: {str(e)}")
            finally:
                driver.quit()

        content.append(
            {
                "type": "text",
                "text": "</website_scrape>Please follow the system prompt step by step and write it out using XML compatible tags, analysing in depth and writing a long essay on each XML point. You must include all of the mandatory-assessment-points and the grading."
            }
        )

        messages = [
            {
                "role": "user",
                "content": content
            }
        ]

        response = client.messages.create(
            model="claude-3-haiku-20240307",
            max_tokens=4000,
            temperature=0,
            messages=messages,
            system=system_prompt
        )

        return response.content[0].text


def extract_grade(response):
    print(response)
    match = re.search(r"<grade-equals>(.*?)</grade-equals>", response, re.DOTALL)
    if match:
        return match.group(1).strip()
    return None


def save_output(firm_id, grade, explanation, error=None):
    output = {
        "grade": grade,
        "explanation": explanation,
        "error": error
    }

    os.makedirs("grader/grades/website-grader", exist_ok=True)
    with open(f"grader/grades/website-grader/{firm_id}.json", "w") as file:
        json.dump(output, file, indent=2)


def process_site(site):
    # Extract the domain from the site URL
    domain_parts = site.split("//")[-1].split("/")[0].split(".")
    domain = ".".join(domain_parts[-2:])

    # Drop sites like yahoo, google, linkedin, etc. Tld agnostic check
    if domain_parts[-2] in ["yahoo", "google", "linkedin", "facebook", "twitter"]:
        return None, None, "Invalid domain"

    # Make a request to the ValueSERP API to get the top search results
    params = {
        'api_key': 'REDACTEDD',
        'q': f"site:{domain} OR site:*.{domain}",
        'location': 'Romania',
        'google_domain': 'google.ro',
        'gl': 'ro',
        'hl': 'ro',
        'output': 'json',
        'num': 3
    }

    response = requests.get('https://api.valueserp.com/search', params=params)
    data = response.json()

    # Print
    # print(data)

    # Extract the URLs of the top search results
    organic_results = data['organic_results'] if 'organic_results' in data else []
    urls = [result['link'] for result in organic_results]

    if not urls:
        return None, None, "No valid URLs found"

    response = analyze_website(urls)
    grade = extract_grade(response)
    explanation = response

    return grade, explanation, None


def process_law_firms(merged_data):
    with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        futures = []
        for law_firm in merged_data:
            firm_id = law_firm['lawyer_id']

            # Check if the JSON file for the law firm already exists
            if os.path.exists(f"grader/grades/website-grader/{firm_id}.json"):
                print(f"Skipping already processed law firm: {firm_id}")
                continue

            website = law_firm.get('website')

            if website:
                future = executor.submit(process_site, website)
                future.add_done_callback(lambda f, fid=firm_id: save_result(f, fid))
                futures.append(future)
            else:
                print(f"No website found for law firm: {firm_id}")
                save_output(firm_id, None, None, "No website found")

        concurrent.futures.wait(futures)


def save_result(future, firm_id):
    grade, explanation, error = future.result()
    save_output(firm_id, grade, explanation, error)
    print(F"Saved result for law firm: {firm_id}")


# Load the merged data from merged_data.json
with open("parsed_data/merged_data.json", "r") as file:
    merged_data = json.load(file)

# Process each law firm
process_law_firms(merged_data)